import os
import json
from typing import Optional, Sequence, Union, Dict
from tqdm import tqdm
import logging

import pandas as pd
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score

from adadec.generator import Generator


def _read_problems_bigcodebench(filename: str):
    problems = {}
    with open(filename, 'r', encoding='utf-8') as f:
        for i, line in enumerate(f):
            line = line.strip()
            if not line:
                continue
            try:
                problem = json.loads(line)
            except json.JSONDecodeError:
                continue
            task_id = problem.get("task_id") or problem.get("id") or f"line_{i}"
            problems[task_id] = problem
    return problems


def _train_logistic_and_find_threshold(df: pd.DataFrame,
                                       target_list: Sequence[str],
                                       test_size: float = 0.2,
                                       random_state: int = 42,
                                       lr_max_iter: int = 1000):
    df = df.copy()
    if 'Rank' not in df.columns:
        raise ValueError("input dataframe needs a 'Rank' column (like in your statistics.parquet).")
    df['y'] = (df['Rank'] == 1).astype(int)

    df = df.dropna(subset=target_list)
    for col in target_list:
        df[col] = np.where(np.isfinite(df[col]), df[col], np.nan)
    df = df.dropna(subset=target_list)

    if len(df) == 0:
        return None

    X = df[list(target_list)]
    y = df['y']

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=random_state)

    train_data = pd.concat([X_train, y_train], axis=1)
    pos = train_data[train_data['y'] == 1]
    neg = train_data[train_data['y'] == 0]
    if len(pos) == 0 or len(neg) == 0:
        X_train_balanced = X_train
        y_train_balanced = y_train
    else:
        nmin = min(len(pos), len(neg))
        pos_s = pos.sample(n=nmin, random_state=random_state)
        neg_s = neg.sample(n=nmin, random_state=random_state)
        train_bal = pd.concat([pos_s, neg_s])
        X_train_balanced = train_bal[list(target_list)]
        y_train_balanced = train_bal['y']

    model = LogisticRegression(max_iter=lr_max_iter)
    model.fit(X_train_balanced, y_train_balanced)

    y_pred_proba = model.predict_proba(X_test)[:, 1]

    best_threshold = 0.0
    best_accuracy = -1.0
    for t in np.arange(0.01, 1.00, 0.01):
        y_pred_thresh = (y_pred_proba >= t).astype(int)
        acc = accuracy_score(y_test, y_pred_thresh)
        if acc > best_accuracy:
            best_accuracy = acc
            best_threshold = float(t)

    learned_entropy = None
    if 'Entropy' in X_test.columns:
        try:
            idx = list(X_train.columns).index('Entropy')
            coef_entropy = model.coef_[0][idx]
            intercept = float(model.intercept_[0])
            p = best_threshold
            eps = 1e-12
            p = np.clip(p, eps, 1 - eps)
            z = np.log(p / (1 - p))
            if coef_entropy == 0:
                learned_entropy = None
            else:
                entropy_at_prob = (z - intercept) / coef_entropy
                learned_entropy = float(entropy_at_prob)
        except ValueError:
            learned_entropy = None

    return {
        "model_obj": model,
        "best_threshold": best_threshold,
        "best_accuracy": best_accuracy,
        "learned_entropy": learned_entropy
    }


def prepare_adadec(model: Union[str, object],
                   tokenizer: Optional[object],
                   train_file: str = '',
                   generate_data_output_file: Optional[str] = None,
                   model_name: Optional[str] = None,
                   learned_thresholds_output_file: str = 'data/learned_thresholds.json',
                   target_list: Sequence[str] = ('Entropy',),
                   test_size: float = 0.2,
                   random_state: int = 42,
                   max_problems: Optional[int] = None,
                   verbose: bool = True) -> Dict:
    if os.path.exists(learned_thresholds_output_file):
        with open(learned_thresholds_output_file, 'r', encoding='utf-8') as f:
            _thresh_dict = json.load(f)
        if model_name in _thresh_dict:
            msg = (f"Threshold for model '{model_name}' already exists in "
                   f"'{learned_thresholds_output_file}'. Skipping re-training.")
            logging.warning(msg)
            return {
                "generate_data_file": generate_data_output_file,
                "learned_thresholds_file": learned_thresholds_output_file,
                "learned_entropy": _thresh_dict.get(model_name),
            }
    
    model_obj = model
    tokenizer_obj = tokenizer
    model_name_from_param = model_name or getattr(model_obj, "name", "model")

    if generate_data_output_file is None:
        generate_data_output_file = os.path.join('data', 'gt_guide_data', f"{model_name_from_param}_statistics.parquet")
        os.makedirs(os.path.dirname(generate_data_output_file), exist_ok=True)
    os.makedirs(os.path.dirname(learned_thresholds_output_file), exist_ok=True) if os.path.dirname(learned_thresholds_output_file) else None

    problems = _read_problems_bigcodebench(train_file)
    items = list(problems.items())
    if max_problems:
        items = items[:max_problems]

    generator = Generator(model_obj, tokenizer_obj, model_name=model_name_from_param)
    if verbose:
        iterator = tqdm(items, desc="Generating gt-guide data", ascii=True)
    else:
        iterator = items

    for task_id, problem in iterator:
        prompt = problem.get("prompt")
        ground_truth = problem.get("canonical_solution")
        try:
            generator.generate_base_on_ground_truth(prompt=prompt,
                                                   ground_truth=ground_truth,
                                                   filename=generate_data_output_file)
        except Exception as e:
            if verbose:
                logging.warning(f"failed to generate for task {task_id}: {e}")
            continue

    if not os.path.exists(generate_data_output_file):
        raise FileNotFoundError(f"generate_data_output_file not found: {generate_data_output_file}")

    df = pd.read_parquet(generate_data_output_file)
    train_result = _train_logistic_and_find_threshold(df, target_list, test_size=test_size, random_state=random_state)

    learned_entropy_value = None

    if train_result is not None:
        learned_entropy_value = train_result.get("learned_entropy")

        if os.path.exists(learned_thresholds_output_file):
            with open(learned_thresholds_output_file, 'r', encoding='utf-8') as f:
                thresh_dict = json.load(f)
        else:
            thresh_dict = {}

        thresh_dict[model_name_from_param] = None if learned_entropy_value is None else round(float(learned_entropy_value), 6)
        
        os.makedirs(os.path.dirname(learned_thresholds_output_file), exist_ok=True) if os.path.dirname(learned_thresholds_output_file) else None
        with open(learned_thresholds_output_file, 'w', encoding='utf-8') as f:
            json.dump(thresh_dict, f, indent=4)

    return {
        "generate_data_file": generate_data_output_file,
        "learned_thresholds_file": learned_thresholds_output_file,
        "learned_entropy": learned_entropy_value,
    }



from typing import Optional, Union, List, Dict, Any
import torch

from adadec.generator import Generator


def _ensure_model_and_tokenizer(model: Optional[object], tokenizer: Optional[object], model_name: Optional[str]):
    if tokenizer is None:
        raise ValueError("tokenizer must be provided")
    model_obj = model
    tokenizer_obj = tokenizer
    resolved_name = model_name or getattr(model_obj, "name", "model")
    return model_obj, tokenizer_obj, resolved_name


def generate_adadec(
    model: Optional[object],
    tokenizer: Optional[object] = None,
    prompts: Union[str, List[str]] = "",
    model_name: Optional[str] = None,
    beam_size: int = 1,
    decoding_mode: str = 'AdaFixL',
    entropy_threshold: Union[str, float] = 'Learned',
    max_new_tokens: int = 512,
    lambda_value: float = 1.0,
    lookahead_length: int = 5,
    lookahead_beam_size: int = 3,
    device: Optional[str] = None,
    stop_words_file: str = ""
) -> Dict[str, Any]:
    is_single = isinstance(prompts, str)
    prompt_list = [prompts] if is_single else list(prompts)

    model_obj, tokenizer_obj, resolved_name = _ensure_model_and_tokenizer(model, tokenizer, model_name)

    if device is None:
        device = "cuda" if torch.cuda.is_available() else "cpu"
    model_obj.to(device)
    model_obj.eval()

    gen = Generator(
        model=model_obj,
        tokenizer=tokenizer_obj,
        model_name=resolved_name,
        beam_size=beam_size,
        decoding_mode=decoding_mode,
        entropy_threshold=entropy_threshold,
        stop_words_file=stop_words_file
    )

    results = []
    for prompt in prompt_list:
        try:
            candidates = gen.generate(
                prompt=prompt,
                beam_size=beam_size,
                max_new_tokens=max_new_tokens,
                lambda_value=lambda_value,
                lookahead_length=lookahead_length,
                lookahead_beam_size=lookahead_beam_size,
            )
        except Exception as e:
            results.append({
                "prompt": prompt,
                "error": str(e),
                "candidates": [],
                "tradition_times": getattr(gen, "tradition_times", None),
                "lookahead_times": getattr(gen, "lookahead_times", None),
            })
            continue

        entry = {
            "prompt": prompt,
            "candidates": candidates,
            "tradition_times": gen.tradition_times,
            "lookahead_times": gen.lookahead_times,
        }
        results.append(entry)

    return {
        "model_name": resolved_name,
        "device": device,
        "results": results
    }


import logging
import torch
import pandas as pd
from transformers import PreTrainedModel, PreTrainedTokenizerBase
import os
import json
from typing import Union
import re



class Generator:
    def __init__(
            self,
            model: PreTrainedModel,
            tokenizer: PreTrainedTokenizerBase,
            model_name: str,
            beam_size: int = 3,
            decoding_mode: str = 'Traditional',
            entropy_threshold: Union[str, float]='Learned',
            stop_words_file: str = "data/stop_words.json"
    ):
        self.model = model
        self.tokenizer = tokenizer
        self.model_name = model_name
        self.beam_size = beam_size
        self.tradition_times = 0
        self.lookahead_times = 0
        self.lookahead_beam_size = 3
        self.generation_counter = 0
        self.decoding_mode = decoding_mode
        
        if os.path.exists(stop_words_file):
            with open(stop_words_file, "r", encoding="utf-8") as f:
                raw_stop_words = json.load(f)
            if not isinstance(raw_stop_words, list):
                raise ValueError(f"stop_words.json format error: should be a list, but got {type(raw_stop_words)}")

            self.stop_words = []
            for sw in raw_stop_words:
                try:
                    self.stop_words.append(re.compile(sw, re.MULTILINE))
                except re.error as e:
                    raise ValueError(f"Invalid regular expression: {sw}, error: {e}")
        else:
            logging.warning(f"stop_words file {stop_words_file} not found, using an empty list.")
            self.stop_words = []

        self.entropy_threshold = None
        if decoding_mode == 'Traditional':
            self.entropy_threshold = float('inf')
        elif decoding_mode == 'AdaFixL':
            if entropy_threshold == 'Learned':
                self.entropy_threshold = self._load_learned_threshold("data/learned_thresholds.json")
            else:
                try:
                    self.entropy_threshold = float(entropy_threshold)
                except ValueError:
                    raise ValueError("Entropy threshold must be a number or 'Learned'")
        else:
            raise ValueError(f"Unsupported decoding_mode: expected 'AdaFixL' , got '{self.decoding_mode}'")


    def _load_learned_threshold(self, threshold_file) -> float:
        if not os.path.exists(threshold_file):
            raise FileNotFoundError(f"Entropy threshold file '{threshold_file}' not found.")
        
        with open(threshold_file, 'r') as f:
            threshold_dict = json.load(f)

        if self.model_name not in threshold_dict:
            raise KeyError(f"Model '{self.model_name}' not found in entropy threshold file.")
        
        return threshold_dict[self.model_name]

    def calculate_entropy(self, next_token_logits):
        next_token_probs_exp = torch.nn.functional.softmax(next_token_logits, dim=-1)
        log_probs = torch.nn.functional.log_softmax(next_token_logits, dim=-1)
        return -torch.sum(next_token_probs_exp * log_probs, dim=-1)

    def select_top_beam_scores(self, beam_size, topk_scores, topk_indices, mode):
        
        if mode == 'Traditional':
            if isinstance(topk_scores, list):
                topk_scores = torch.cat(topk_scores, dim=0)
                topk_indices = torch.cat(topk_indices, dim=0)

            total_candidates = topk_scores.size(0)

            if total_candidates == beam_size:
                selected_groups = torch.arange(beam_size, dtype=torch.long, device=topk_scores.device)
                return topk_scores, topk_indices, selected_groups

            elif total_candidates == beam_size * beam_size:
                final_scores, flat_indices = torch.topk(topk_scores, beam_size)  # [beam_size]

                selected_groups = flat_indices // beam_size
                token_pos_in_group = flat_indices % beam_size

                final_indices = []
                for group, pos in zip(selected_groups, token_pos_in_group):
                    index = group * beam_size + pos
                    final_indices.append(topk_indices[index])
                final_indices = torch.stack(final_indices)

                return final_scores, final_indices, selected_groups

            else:
                raise ValueError(f"Unsupported topk_scores size: expected {beam_size} or {beam_size * beam_size}, got {total_candidates}")
        
        elif mode == 'AdaFixL':
            total_candidates = topk_scores.size(0)
            assert total_candidates % beam_size == 0, "topk_scores size must be divisible by beam_size"
            
            batch_size = total_candidates // beam_size

            # reshape to [batch_size, beam_size]
            topk_scores = topk_scores.view(batch_size, beam_size)
            topk_indices = topk_indices.view(batch_size, beam_size)

            final_scores, local_indices = torch.topk(topk_scores, beam_size, dim=-1)  # [batch_size, beam_size]

            batch_indices = torch.arange(batch_size).unsqueeze(1).to(topk_indices.device)  # [batch_size, 1]
            final_indices = topk_indices[batch_indices, local_indices]  # [batch_size, beam_size]

            selected_groups = batch_indices.expand(-1, beam_size)

            return final_scores, final_indices, selected_groups
        
        else:
           raise ValueError(f"Unsupported decoding_mode: got {self.decoding_mode}")

    def scoring_function(self, next_token_logits, beam_scores, beam_size):
        log_probs = torch.nn.functional.log_softmax(next_token_logits, dim=-1)  # [batch, vocab]
        topk_scores, topk_indices = torch.topk(log_probs, beam_size, dim=-1)    # [batch, beam_size]

        if beam_scores.dim() == 1:
            beam_scores = beam_scores.unsqueeze(1).expand(-1, beam_size)

        topk_scores = topk_scores + beam_scores

        topk_scores = topk_scores.view(-1)
        topk_indices = topk_indices.view(-1)

        return topk_scores, topk_indices
    
    def generate(
            self,
            prompt,
            beam_size=1,
            max_new_tokens=512,
            lambda_value=1.0,
            lookahead_length=5,
            lookahead_beam_size=3,
    ):
        self.beam_size = beam_size
        self.lookahead_beam_size = lookahead_beam_size
        
        self.tradition_times = 0
        self.lookahead_times = 0

        token_ids = self.tokenizer([prompt], add_special_tokens=True, padding=True, truncation=True,
                                   return_tensors="pt").input_ids
        token_ids = token_ids.to(self.model.device)
        
        decoded_prompt = self.tokenizer.decode(
            self.tokenizer(prompt, return_tensors="pt").input_ids[0].to(self.model.device),
            skip_special_tokens=True
        )

        token_ids = token_ids.repeat(beam_size, 1)
        attention_mask = torch.ones_like(token_ids).to(self.model.device)

        beam_scores = torch.zeros(beam_size, dtype=torch.float).to(self.model.device)

        is_finished = [False] * beam_size

        beam_indices = torch.zeros(beam_size, dtype=torch.long, device=self.model.device)

        with torch.no_grad():
            for step in range(max_new_tokens):
                if all(is_finished):
                    break

                outputs = self.model(input_ids=token_ids, attention_mask=attention_mask)
                next_token_logits = outputs.logits[:, -1, :]

                topk_k_scores = []
                topk_k_indices = []

                entropy = self.calculate_entropy(next_token_logits)

                for i in range(self.beam_size):
                    if entropy[i] < self.entropy_threshold or is_finished[i]:
                        curr_topk_scores, curr_topk_indices = self.scoring_function(next_token_logits[i],
                                                                                    beam_scores[i],
                                                                                    beam_size=beam_size)
                        topk_k_scores.append(curr_topk_scores)
                        topk_k_indices.append(curr_topk_indices)
                        self.tradition_times += 1
                    else:
                        curr_topk_scores, curr_topk_indices = self.lookahead_scoring_function(
                            decoded_prompt,
                            next_token_logits[i],
                            token_ids[i],
                            beam_scores[i],
                            lookahead_length=lookahead_length,
                            lambda_value=lambda_value,
                        )
                        topk_k_scores.append(curr_topk_scores)
                        topk_k_indices.append(curr_topk_indices)
                        self.lookahead_times += 1

                if not topk_k_scores:
                    break

                if step == 0:
                    topk_scores = topk_k_scores[0]
                    topk_indices = topk_k_indices[0]
                else:
                    topk_scores, topk_indices, beam_indices = self.select_top_beam_scores(
                        beam_size=beam_size,
                        topk_scores=topk_k_scores,
                        topk_indices=topk_k_indices,
                        mode='Traditional'
                    )

                token_indices = topk_indices % next_token_logits.shape[-1]

                token_ids = torch.cat([
                    token_ids[beam_indices],
                    token_indices.unsqueeze(-1)
                ], dim=-1)
                beam_scores = topk_scores

                for j in range(beam_size):
                    if token_indices[j] == self.tokenizer.eos_token_id:
                        is_finished[j] = True
                        continue

                    prompt_ids = self.tokenizer(prompt, return_tensors="pt").input_ids[0].to(self.model.device)
                    decoded_prompt = self.tokenizer.decode(prompt_ids, skip_special_tokens=True)
                    decoded_seq = self.tokenizer.decode(token_ids[j], skip_special_tokens=True)
                    current_sequence = decoded_seq[len(decoded_prompt):]
                    
                    is_stop = any(pattern.search(current_sequence) for pattern in self.stop_words)

                    if is_stop:
                        is_finished[j] = True

                attention_mask = token_ids.ne(self.tokenizer.pad_token_id).to(self.model.device)

                torch.cuda.empty_cache()

        decoded_sequences = [
            self.tokenizer.decode(seq, skip_special_tokens=True)
            for seq in token_ids
        ]
        

        torch.cuda.empty_cache()

        return [gen[len(prompt):] for gen in decoded_sequences]

    def lookahead_scoring_function(self, decoded_prompt, next_token_logits, token_ids, beam_scores,
                                lookahead_length, lambda_value):
        lookahead_beam_size = self.lookahead_beam_size
        beam_size = self.beam_size

        history_topk_score = beam_scores

        topk_scores, topk_indices = self.scoring_function(
            next_token_logits, beam_scores, beam_size=lookahead_beam_size)

        current_topk_scores = topk_scores - history_topk_score
        token_indices = topk_indices % next_token_logits.shape[-1]

        token_ids = token_ids.repeat(lookahead_beam_size, 1)
        token_ids = torch.cat([token_ids, token_indices.unsqueeze(-1)], dim=-1)

        lookahead_scores, actual_lengths = self.get_lookahead_score(
            token_ids=token_ids,
            lookahead_length=lookahead_length,
            decoded_prompt=decoded_prompt
        )
        total_scores = history_topk_score + (current_topk_scores + lookahead_scores) / (actual_lengths + 1) * lambda_value

        topk_scores, topk_indices_temp = torch.topk(total_scores, beam_size)
        topk_indices = topk_indices[topk_indices_temp]

        return topk_scores, topk_indices

    def get_lookahead_score(self, token_ids, lookahead_length, decoded_prompt):
        batch_size = token_ids.shape[0]
        beam_size = self.beam_size

        device = self.model.device

        token_ids = token_ids.repeat_interleave(beam_size, dim=0)
        attention_mask = torch.ones_like(token_ids).to(device)

        lookahead_scores = torch.zeros(token_ids.size(0), dtype=torch.float).to(device)
        is_finished = torch.zeros(token_ids.size(0), dtype=torch.bool).to(device)
        actual_lookahead_length = torch.zeros(token_ids.size(0), dtype=torch.long).to(device)

        with torch.no_grad():
            for _ in range(lookahead_length):
                if is_finished.all():
                    break

                outputs = self.model(input_ids=token_ids, attention_mask=attention_mask)
                next_token_logits = outputs.logits[:, -1, :]

                topk_scores, topk_indices = self.scoring_function(next_token_logits, lookahead_scores, beam_size=beam_size)

                selected_scores, selected_indices, beam_idx = self.select_top_beam_scores(
                    beam_size=beam_size,
                    topk_scores=topk_scores,
                    topk_indices=topk_indices,
                    mode='AdaFixL'
                )

                lookahead_scores = selected_scores
                actual_lookahead_length += (~is_finished).long()
                next_tokens = selected_indices % next_token_logits.shape[-1]

                token_ids = torch.cat([
                    token_ids[beam_idx],
                    next_tokens.unsqueeze(-1)
                ], dim=-1)
                
                if token_ids.ndim == 1:
                    token_ids = token_ids.unsqueeze(0) 
                else:
                    token_ids = token_ids.view(-1, token_ids.shape[-1])

                decoded_seqs = self.tokenizer.batch_decode(token_ids.tolist(), skip_special_tokens=True)

                for i in range(token_ids.size(0)):
                    if is_finished[i]:
                        continue
                    current_seq = decoded_seqs[i][len(decoded_prompt):]
                    lines = current_seq.split('\n')
                    is_consecutive_empty = len(lines) >= 4 and all(line.strip() == '' for line in lines[-4:])
                    all_lines_valid = all(not line or line.startswith((' ', '\t')) for line in lines)
                    is_finished[i] = is_consecutive_empty or not all_lines_valid

                attention_mask = token_ids.ne(self.tokenizer.pad_token_id).to(device)

        lookahead_scores = lookahead_scores.view(batch_size, beam_size)
        actual_lookahead_length = actual_lookahead_length.view(batch_size, beam_size)

        max_scores, _ = lookahead_scores.max(dim=1)
        max_lengths, _ = actual_lookahead_length.max(dim=1)

        return max_scores, max_lengths

    def generate_base_on_ground_truth(
            self,
            prompt,
            ground_truth,
            filename,
    ):
        generation_id = self.generation_counter
        self.generation_counter += 1

        device = next(self.model.parameters()).device
        prompt_ids = self.tokenizer(prompt, return_tensors="pt").input_ids.to(device)
        gt_ids = self.tokenizer(ground_truth, return_tensors="pt").input_ids[0].to(device)
        current_sequence = prompt_ids

        log_data = []

        for i, gt_token in enumerate(gt_ids):
            with torch.no_grad():
                outputs = self.model(current_sequence)
                logits = outputs.logits
                next_logits = logits[0, -1, :]
            # topk_values, topk_indices = torch.topk(next_logits, k=50)
            entropy = self.calculate_entropy(next_logits)
            sorted_logits, sorted_indices = torch.sort(next_logits, descending=True)
            rank_tensor = (sorted_indices == gt_token).nonzero()
            rank = rank_tensor.item() + 1 if rank_tensor.numel() > 0 else -1

            log_data.append({
                "GenerationID": generation_id,
                "Iteration": i,
                "Entropy": float(entropy),
                "GroundTruthToken": gt_token.item(),
                "Rank": rank,
                # "Top50_TokenIDs": topk_indices.tolist(),
                # "Top50_TokenLogits": topk_values.tolist()
            })

            gt_token_id = gt_token.unsqueeze(0).unsqueeze(0).to(device)
            current_sequence = torch.cat([current_sequence, gt_token_id], dim=1)

            torch.cuda.empty_cache()

        new_df = pd.DataFrame(log_data)

        if os.path.exists(filename):
            old_df = pd.read_parquet(filename, engine="pyarrow")
            combined_df = pd.concat([old_df, new_df], ignore_index=True)
            combined_df.to_parquet(filename, engine="pyarrow", index=False)
        else:
            new_df.to_parquet(filename, engine="pyarrow", index=False)